from typing import List, Tuple, Union

import numpy as np
import pandas as pd
from napari.layers import Points

from surforama.constants import (
    NAPARI_NORMAL_0,
    NAPARI_NORMAL_1,
    NAPARI_NORMAL_2,
    NAPARI_UP_0,
    NAPARI_UP_1,
    NAPARI_UP_2,
    ROTATION,
)
from surforama.utils.geometry import rotate_around_vector


def vectors_data_from_points_layer(
    points_layer: Points,
) -> Tuple[np.ndarray, np.ndarray]:
    """Get the vectors data for normal and up vectors from an oriented points layer.

    Parameters
    ----------
    points_layer : Points
        The oriented points layer from which to get the vectors data.

    Returns
    -------
    normals_data : np.ndarray
        The Vectors layer data for the normal vectors
    up_data : np.ndarray
        The Vectors layer data for the up vectors
    """

    # get the point data
    point_coordinates = points_layer.data
    features_table = points_layer.features
    return vectors_data_from_points_data(
        point_coordinates=point_coordinates, features_table=features_table
    )


def vectors_data_from_points_data(
    point_coordinates: np.ndarray, features_table: pd.DataFrame
) -> Tuple[np.ndarray, np.ndarray]:
    """Get the vectors data for normal and up vectors from an oriented points layer.

    Parameters
    ----------
    point_coordinates : np.ndarray
        (n, 3) array of the coordinates of each point.
    features_table : pd.DataFrame
        The layer features table.

    Returns
    -------
    normals_data : np.ndarray
        The Vectors layer data for the normal vectors
    up_data : np.ndarray
        The Vectors layer data for the up vectors
    """
    # check in the points layer has any data
    if len(point_coordinates) == 0:
        # if there are no points, there are no vectors
        return np.empty((0, 2, 3)), np.empty((0, 2, 3))
    # get the vectors
    normal_vectors = features_table[
        [NAPARI_NORMAL_0, NAPARI_NORMAL_1, NAPARI_NORMAL_2]
    ].to_numpy()
    up_vectors = features_table[
        [NAPARI_UP_0, NAPARI_UP_1, NAPARI_UP_2]
    ].to_numpy()
    rotations = features_table[ROTATION].to_numpy()
    rotated_up_vectors = rotate_around_vector(
        rotate_around=normal_vectors, to_rotate=up_vectors, angle=rotations
    )

    # make the vector data
    n_points = point_coordinates.shape[0]
    normal_vector_data = np.zeros((n_points, 2, 3))
    normal_vector_data[:, 0, :] = point_coordinates
    normal_vector_data[:, 1, :] = normal_vectors

    up_vector_data = np.zeros((n_points, 2, 3))
    up_vector_data[:, 0, :] = point_coordinates
    up_vector_data[:, 1, :] = rotated_up_vectors

    return normal_vector_data, up_vector_data


def get_vectors_data_from_points_layer(
    points_layer: Points, point_index: Union[int, List[int]]
) -> Tuple[np.ndarray, np.ndarray, Union[float, np.ndarray]]:
    """Get the point vectors from the points features table.

    Parameters
    ----------
    points_layer : Points
        The points layer to get the vectors from.
    point_index : Union[int, List[int]]
        The indices of the points to get the vectors for.

    Returns
    -------
    selected_normal_vectors : np.ndarray
        The normal vectors from the selected points.
    selected_up_vectors : np.ndarray
        The up-vectors from the selected points.
    selected_rotations : np.ndarray
        The rotations from the selected points.
    """
    feature_table = points_layer.features

    # get the normal vector
    normal_vectors = feature_table[
        [NAPARI_NORMAL_0, NAPARI_NORMAL_1, NAPARI_NORMAL_2]
    ].to_numpy()
    selected_normal_vector = normal_vectors[point_index]

    # get the up vector
    up_vectors = feature_table[
        [NAPARI_UP_0, NAPARI_UP_1, NAPARI_UP_2]
    ].to_numpy()
    selected_up_vector = up_vectors[point_index]

    # get the rotation
    rotations = feature_table[ROTATION].to_numpy()
    selected_rotations = rotations[point_index]

    return selected_normal_vector, selected_up_vector, selected_rotations


def update_vectors_data_on_points_layer(
    points_layer: Points,
    point_index: Union[int, List[int]],
    normal_vectors: np.ndarray,
    up_vectors: np.ndarray,
    rotations: np.ndarray,
):
    """Update all fields in the vectors data on a points layer.

    Parameters
    ----------
    points_layer : Points
        The points layer to be updated.
    point_index : Union[int, List[int]]
        The indices of the points to be updated.
    normal_vectors : np.ndarray
        (n, 3) array of the new normal vectors.
    up_vectors : np.ndarray
        (n, 3) array of the new up vectors.
    rotations : np.ndarray
        (n,) array of the new rotation angles in radians.
    """
    if isinstance(point_index, int):
        point_index = [point_index]

    feature_table = points_layer.features

    # get the normal vector
    current_normal_vectors = feature_table[
        [NAPARI_NORMAL_0, NAPARI_NORMAL_1, NAPARI_NORMAL_2]
    ].to_numpy()
    current_normal_vectors[point_index, :] = normal_vectors

    # get the up vector
    current_up_vectors = feature_table[
        [NAPARI_UP_0, NAPARI_UP_1, NAPARI_UP_2]
    ].to_numpy()
    current_up_vectors[point_index, :] = up_vectors

    # get the rotation
    current_rotations = feature_table[ROTATION].to_numpy()
    current_rotations[point_index] = rotations

    # set the values
    feature_table[NAPARI_NORMAL_0] = current_normal_vectors[:, 0]
    feature_table[NAPARI_NORMAL_1] = current_normal_vectors[:, 1]
    feature_table[NAPARI_NORMAL_2] = current_normal_vectors[:, 2]
    feature_table[NAPARI_UP_0] = current_up_vectors[:, 0]
    feature_table[NAPARI_UP_1] = current_up_vectors[:, 1]
    feature_table[NAPARI_UP_2] = current_up_vectors[:, 2]
    feature_table[ROTATION] = current_rotations


def update_rotations_on_points_layer(
    points_layer: Points,
    point_index: Union[int, List[int]],
    rotations: np.ndarray,
) -> None:
    """Update the rotations for selected points.

    Parameters
    ----------
    points_layer : Points
        The points layer to be updated.
    point_index : Union[int, List[int]]
        The indices of the points to update the rotations on.
    rotations : np.ndarray
        The new rotations to set.
    """
    if isinstance(point_index, int):
        point_index = [point_index]

    feature_table = points_layer.features

    # get the rotation
    current_rotations = feature_table[ROTATION].to_numpy()
    current_rotations[point_index] = rotations

    # set the values
    feature_table[ROTATION] = current_rotations



# Relion star table names for particle coordinates
STAR_X_COLUMN_NAME = "rlnCoordinateX"
STAR_Y_COLUMN_NAME = "rlnCoordinateY"
STAR_Z_COLUMN_NAME = "rlnCoordinateZ"
STAR_ROTATION_0 = "rlnAngleRot"
STAR_ROTATION_1 = "rlnAngleTilt"
STAR_ROTATION_2 = "rlnAnglePsi"


# column names for napari points layer for oriented particles
NAPARI_NORMAL_0 = "normal-0"
NAPARI_NORMAL_1 = "normal-1"
NAPARI_NORMAL_2 = "normal-2"
NAPARI_UP_0 = "up-0"
NAPARI_UP_1 = "up-1"
NAPARI_UP_2 = "up-2"
ROTATION = "rotation"



import numpy as np

from surforama.utils.geometry import rotate_around_vector


def test_rotate_around_axis_single_vector():
    rotate_around = np.array([1, 0, 0])
    to_rotate = np.array([0, 1, 0])

    rotation_amount = np.pi / 2
    rotated_vector = rotate_around_vector(
        rotate_around=rotate_around, to_rotate=to_rotate, angle=rotation_amount
    )

    assert np.allclose(rotated_vector, np.array([0, 0, 1]))


def test_rotate_around_axis_multiple_vectors():
    rotate_around = np.array([[1, 0, 0], [0, 1, 0]])
    to_rotate = np.array([[0, 1, 0], [0, 0, 1]])

    rotation_amount = np.pi / 2
    rotated_vector = rotate_around_vector(
        rotate_around=rotate_around, to_rotate=to_rotate, angle=rotation_amount
    )

    assert np.allclose(rotated_vector, np.array([[0, 0, 1], [1, 0, 0]]))


def test_rotate_around_axis_multiple_vectors_multiple_angles():
    rotate_around = np.array([[1, 0, 0], [0, 1, 0]])
    to_rotate = np.array([[0, 1, 0], [0, 0, 1]])

    rotation_amount = np.array([np.pi / 2, np.pi])
    rotated_vector = rotate_around_vector(
        rotate_around=rotate_around, to_rotate=to_rotate, angle=rotation_amount
    )

    assert np.allclose(rotated_vector, np.array([[0, 0, 1], [0, 0, -1]]))



import numpy as np
import pandas as pd
from napari.layers import Points

from surforama.constants import (
    NAPARI_NORMAL_0,
    NAPARI_NORMAL_1,
    NAPARI_NORMAL_2,
    NAPARI_UP_0,
    NAPARI_UP_1,
    NAPARI_UP_2,
    ROTATION,
)
from surforama.io.star import (
    load_points_layer_data_from_star_file,
    oriented_points_to_star_file,
)
from surforama.utils.geometry import rotate_around_vector


def test_star_file_round_trip(tmp_path):
    """Test saving and loading a star file is lossless."""
    # make a points layer with an oriented points feature table
    point_coordinates = np.array([[10, 10, 10], [20, 20, 20]])
    normal_vectors = np.array([[1, 0, 0], [1, 0, 0]])
    up_vectors = np.array([[0, 0, 1], [0, 1, 0]])
    rotations = np.array([0, np.pi / 2])
    feature_table = pd.DataFrame(
        {
            NAPARI_NORMAL_0: normal_vectors[:, 0],
            NAPARI_NORMAL_1: normal_vectors[:, 1],
            NAPARI_NORMAL_2: normal_vectors[:, 2],
            NAPARI_UP_0: up_vectors[:, 0],
            NAPARI_UP_1: up_vectors[:, 1],
            NAPARI_UP_2: up_vectors[:, 2],
            ROTATION: rotations,
        }
    )
    points_layer = Points(point_coordinates, features=feature_table)

    # save the feature table to a star file
    star_file_path = tmp_path / "test.star"
    oriented_points_to_star_file(points_layer, output_path=star_file_path)

    # load the star file
    returned_points_coordinates, returned_features_table = (
        load_points_layer_data_from_star_file(star_file_path)
    )

    # compare the point coordinates
    np.testing.assert_allclose(returned_points_coordinates, point_coordinates)

    # compare the normal vectors
    returned_normal_vectors = returned_features_table[
        [NAPARI_NORMAL_0, NAPARI_NORMAL_1, NAPARI_NORMAL_2]
    ]
    np.testing.assert_allclose(returned_normal_vectors, normal_vectors)

    # compare the up vectors
    expected_up_vectors = rotate_around_vector(
        rotate_around=normal_vectors, to_rotate=up_vectors, angle=rotations
    )
    returned_up_vectors = returned_features_table[
        [NAPARI_UP_0, NAPARI_UP_1, NAPARI_UP_2]
    ].to_numpy()
    np.testing.assert_allclose(
        returned_up_vectors, expected_up_vectors, atol=1e-15
    )



from pathlib import Path
from typing import TYPE_CHECKING, Optional

from magicgui import magicgui
from napari.utils.notifications import show_warning
from qtpy.QtWidgets import QGroupBox, QTabWidget, QVBoxLayout, QWidget

if TYPE_CHECKING:
    from surforama.gui.qt_surface_picker import QtSurfacePicker
from surforama.io.star import (
    load_points_layer_data_from_star_file,
    oriented_points_to_star_file,
)
from surforama.utils.napari import vectors_data_from_points_data


class QtPointIO(QGroupBox):
    def __init__(
        self,
        surface_picker: "QtSurfacePicker",
        parent: Optional[QWidget] = None,
    ):
        super().__init__("Save points", parent=parent)
        self.surface_picker = surface_picker

        # make the points saving widget
        self.file_saving_widget = magicgui(
            self._write_star_file,
            file_path={"mode": "w"},
            call_button="Save to star file",
        )
        self.file_loading_widget = magicgui(
            self._load_star_file,
            file_path={"mode": "r"},
            call_button="Load from star file",
        )

        # make the tab widget
        self.tab_widget = QTabWidget(self)
        self.tab_widget.addTab(self.file_saving_widget.native, "save")
        self.tab_widget.addTab(self.file_loading_widget.native, "load")

        # make the layout
        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.tab_widget)

    def _write_star_file(self, file_path: Path):
        oriented_points_to_star_file(
            points_layer=self.surface_picker.points_layer,
            output_path=file_path,
        )

    def _load_star_file(self, file_path: Path):
        """Load oriented points from a star file and add them
        to the viewer.
        """
        if self.surface_picker.enabled is False:
            show_warning("The surface picker must be enabled to load a file.")
            return
        # get the points data
        point_coordinates, features_table = (
            load_points_layer_data_from_star_file(file_path=file_path)
        )

        # get the vectors data
        normal_data, up_data = vectors_data_from_points_data(
            point_coordinates=point_coordinates, features_table=features_table
        )

        # add the data to the viewer
        self.surface_picker.points_layer.data = point_coordinates
        self.surface_picker.points_layer.features = features_table

        self.surface_picker.normal_vectors_layer.data = normal_data
        self.surface_picker.up_vectors_layer.data = up_data

        self.surface_picker.normal_vectors_layer.edge_color = "purple"
        self.surface_picker.up_vectors_layer.edge_color = "orange"



from pathlib import Path
from typing import Tuple, Union

import numpy as np
import pandas as pd
import starfile
from napari.layers import Points
from scipy.spatial.transform import Rotation as R

from surforama.constants import (
    NAPARI_NORMAL_0,
    NAPARI_NORMAL_1,
    NAPARI_NORMAL_2,
    NAPARI_UP_0,
    NAPARI_UP_1,
    NAPARI_UP_2,
    ROTATION,
    STAR_ROTATION_0,
    STAR_ROTATION_1,
    STAR_ROTATION_2,
    STAR_X_COLUMN_NAME,
    STAR_Y_COLUMN_NAME,
    STAR_Z_COLUMN_NAME,
)
from surforama.utils.geometry import rotate_around_vector


def load_points_from_star_table(star_table: pd.DataFrame) -> np.ndarray:
    """Get point coordinates from a Relion-formatted star table.

    Currently this does not account for shifts.

    Parameters
    ----------
    star_table : pd.DataFrame
        The table from which to extract the point coordinates
    """
    return star_table[
        [STAR_Z_COLUMN_NAME, STAR_Y_COLUMN_NAME, STAR_X_COLUMN_NAME]
    ].to_numpy()


def load_orientations_from_star_table(star_table: pd.DataFrame):
    """Get orientations from a Relion-formatted star table."""
    eulers = star_table[
        ["rlnAngleRot", "rlnAngleTilt", "rlnAnglePsi"]
    ].to_numpy()
    return R.from_euler(seq="ZYZ", angles=eulers, degrees=True).inv()


def load_points_layer_data_from_star_file(
    file_path: str,
) -> Tuple[np.ndarray, pd.DataFrame]:
    """Load an oriented Points layer from a Relion-formatted star file"""
    star_table = starfile.read(file_path)
    point_coordinates = load_points_from_star_table(star_table)
    orientations = load_orientations_from_star_table(star_table)

    # get the normal vectors
    initial_normal_vector = np.array([0, 0, 1])
    normal_vectors = orientations.apply(initial_normal_vector)[:, ::-1]

    # get the up vectors
    initial_up_vector = np.array([0, 1, 0])
    up_vectors = orientations.apply(initial_up_vector)[:, ::-1]

    # initialize the rotations
    n_points = point_coordinates.shape[0]
    rotations = np.zeros((n_points,))

    feature_table = pd.DataFrame(
        {
            NAPARI_NORMAL_0: normal_vectors[:, 0],
            NAPARI_NORMAL_1: normal_vectors[:, 1],
            NAPARI_NORMAL_2: normal_vectors[:, 2],
            NAPARI_UP_0: up_vectors[:, 0],
            NAPARI_UP_1: up_vectors[:, 1],
            NAPARI_UP_2: up_vectors[:, 2],
            ROTATION: rotations,
        }
    )

    return point_coordinates, feature_table


def oriented_points_to_star_table(points_layer: Points):
    """points with orientations to a star-formatted DataFrame."""
    # get the point coordinates
    napari_coordinates = points_layer.data
    relion_coordinates = napari_coordinates[:, ::-1]

    # get the point orientations
    features_table = points_layer.features
    normal_vectors = features_table[
        [NAPARI_NORMAL_0, NAPARI_NORMAL_1, NAPARI_NORMAL_2]
    ].to_numpy()
    up_vectors = features_table[
        [NAPARI_UP_0, NAPARI_UP_1, NAPARI_UP_2]
    ].to_numpy()
    rotations = features_table[ROTATION].to_numpy()
    rotated_up_vectors = rotate_around_vector(
        rotate_around=normal_vectors, to_rotate=up_vectors, angle=rotations
    )
    third_basis = np.cross(normal_vectors, rotated_up_vectors)
    n_points = napari_coordinates.shape[0]
    particle_orientations = np.zeros((n_points, 3, 3))
    particle_orientations[:, :, 0] = third_basis[:, ::-1]
    particle_orientations[:, :, 1] = rotated_up_vectors[:, ::-1]
    particle_orientations[:, :, 2] = normal_vectors[:, ::-1]
    euler_angles = (
        R.from_matrix(particle_orientations)
        .inv()
        .as_euler(seq="ZYZ", degrees=True)
    )

    return pd.DataFrame(
        {
            STAR_X_COLUMN_NAME: relion_coordinates[:, 0],
            STAR_Y_COLUMN_NAME: relion_coordinates[:, 1],
            STAR_Z_COLUMN_NAME: relion_coordinates[:, 2],
            STAR_ROTATION_0: euler_angles[:, 0],
            STAR_ROTATION_1: euler_angles[:, 1],
            STAR_ROTATION_2: euler_angles[:, 2],
        }
    )


def oriented_points_to_star_file(
    points_layer: Points, output_path: Union[str, Path]
):
    """points with orientations to a star file"""
    star_table = oriented_points_to_star_table(points_layer)
    starfile.write(star_table, output_path)



from typing import List, Optional

import mrcfile
import napari
import numpy as np
import trimesh
from magicgui import magicgui
from napari.layers import Image, Surface
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QHBoxLayout,
    QLabel,
    QSlider,
    QVBoxLayout,
    QWidget,
)
from scipy.ndimage import map_coordinates

from surforama.gui.qt_mesh_generator import QtMeshGenerator
from surforama.gui.qt_point_io import QtPointIO
from surforama.gui.qt_surface_picker import QtSurfacePicker
from surforama.io import read_obj_file


class QtSurforama(QWidget):
    def __init__(
        self,
        viewer: napari.Viewer,
        surface_layer: Optional[Surface] = None,
        volume_layer: Optional[Image] = None,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent=parent)
        self.viewer = viewer
        self._enabled = False

        # make the layer selection widget
        self._layer_selection_widget = magicgui(
            self._set_layers,
            surface_layer={"choices": self._get_valid_surface_layers},
            image_layer={"choices": self._get_valid_image_layers},
            call_button="start surfing",
        )

        # add callback to update choices
        self.viewer.layers.events.inserted.connect(self._on_layer_update)
        self.viewer.layers.events.removed.connect(self._on_layer_update)

        # make the slider to change thickness
        self.slider = QSlider()
        self.slider.setOrientation(Qt.Horizontal)
        self.slider.setMinimum(-100)
        self.slider.setMaximum(100)
        self.slider.setValue(0)
        self.slider.valueChanged.connect(self.slide_points)
        self.slider.setVisible(False)
        self.slider_value = QLabel("0 vx", self)
        self.slider_value.setVisible(False)
        self.slider_title = QLabel("Extend/contract surface")
        self.slider_title.setVisible(False)

        self.sliderLayout = QHBoxLayout()
        self.sliderLayout.addWidget(self.slider)
        self.sliderLayout.addWidget(self.slider_value)

        # New slider for sampling depth

        self.sampling_depth_slider = QSlider()
        self.sampling_depth_slider.setOrientation(Qt.Horizontal)
        self.sampling_depth_slider.setMinimum(1)
        self.sampling_depth_slider.setMaximum(100)
        self.sampling_depth_slider.setValue(10)
        self.sampling_depth_slider.valueChanged.connect(
            self.update_colors_based_on_sampling
        )
        self.sampling_depth_slider.setVisible(False)
        self.sampling_depth_value = QLabel("10", self)
        self.sampling_depth_value.setVisible(False)
        self.sampling_depth_title = QLabel("Surface Thickness")
        self.sampling_depth_title.setVisible(False)

        self.sampling_depth_sliderLayout = QHBoxLayout()
        self.sampling_depth_sliderLayout.addWidget(self.sampling_depth_slider)
        self.sampling_depth_sliderLayout.addWidget(self.sampling_depth_value)

        # make the picking widget
        self.picking_widget = QtSurfacePicker(surforama=self, parent=self)
        self.picking_widget.setVisible(False)

        # make the saving widget
        self.point_writer_widget = QtPointIO(
            surface_picker=self.picking_widget, parent=self
        )
        self.point_writer_widget.setVisible(False)

        self.mesh_generator_widget = QtMeshGenerator(viewer, parent=self)

        # make the layout
        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.mesh_generator_widget)
        self.layout().addWidget(self._layer_selection_widget.native)
        # self.layout().addWidget(QLabel("Extend/contract surface"))
        self.layout().addWidget(self.slider_title)
        self.layout().addLayout(self.sliderLayout)
        # self.layout().addWidget(QLabel("Surface Thickness"))
        self.layout().addWidget(self.sampling_depth_title)
        self.layout().addLayout(self.sampling_depth_sliderLayout)
        self.layout().addWidget(self.picking_widget)
        self.layout().addWidget(self.point_writer_widget)
        self.layout().addStretch()

        # set the layers
        self._set_layers(surface_layer=surface_layer, image_layer=volume_layer)

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, enabled: bool):
        if enabled == self.enabled:
            # no change
            return

        if enabled:
            # make the widgets visible
            self.slider.setVisible(True)
            self.sampling_depth_slider.setVisible(True)
            self.picking_widget.setVisible(True)
            self.point_writer_widget.setVisible(True)
            self.slider_title.setVisible(True)
            self.slider_value.setVisible(True)
            self.sampling_depth_title.setVisible(True)
            self.sampling_depth_value.setVisible(True)
        else:
            # make the widgets visible
            self.slider.setVisible(False)
            self.sampling_depth_slider.setVisible(False)
            self.picking_widget.setVisible(False)
            self.point_writer_widget.setVisible(False)
            self.slider_title.setVisible(False)
            self.slider_value.setVisible(False)
            self.sampling_depth_title.setVisible(False)
            self.sampling_depth_value.setVisible(False)

        self._enabled = enabled

    def _set_layers(
        self,
        surface_layer: Surface,
        image_layer: Image,
    ):
        self.surface_layer = surface_layer
        self.image_layer = image_layer

        if (surface_layer is None) or (image_layer is None):
            return

        # Create a mesh object using trimesh
        self.mesh = trimesh.Trimesh(
            vertices=surface_layer.data[0], faces=surface_layer.data[1]
        )

        self.vertices = self.mesh.vertices
        self.faces = self.mesh.faces

        # Compute vertex normals
        self.color_values = np.ones((self.mesh.vertices.shape[0],))

        self.surface_layer.data = (
            self.vertices,
            self.faces,
            self.color_values,
        )
        self.surface_layer.shading = "none"
        self.surface_layer.refresh()

        self.normals = self.mesh.vertex_normals
        self.volume = image_layer.data.astype(np.float32)

        self.enabled = True

    def _get_valid_surface_layers(self, combo_box) -> List[Surface]:
        return [
            layer
            for layer in self._viewer.layers
            if isinstance(layer, napari.layers.Surface)
        ]

    def _on_layer_update(self, event=None):
        """When the model updates the selected layer, update widgets."""
        self._layer_selection_widget.reset_choices()

        # check if the stored layers are still around
        layer_deleted = False
        if (
            self.surface_layer is not None
        ) and self.surface_layer not in self.viewer.layers:
            # remove the surface layer if it has been deleted.
            self.surface_layer = None
            layer_deleted = True

        if (self.image_layer is not None) and (
            self.image_layer not in self.viewer.layers
        ):
            # remove the surface layer if it has been deleted.
            self.image_layer = None
            layer_deleted = True

        if layer_deleted:
            self.enabled = False

    def _get_valid_image_layers(self, combo_box) -> List[Image]:
        return [
            layer
            for layer in self._viewer.layers
            if isinstance(layer, napari.layers.Image)
        ]

    def get_point_colors(self, points):
        point_values = map_coordinates(
            self.volume, points.T, order=1, mode="nearest"
        )

        normalized_values = (point_values - point_values.min()) / (
            point_values.max() - point_values.min() + np.finfo(float).eps
        )
        return normalized_values

    def get_point_set(self):
        return self.mesh.vertices

    def get_faces(self):
        return self.mesh.faces

    def slide_points(self, value):
        # Calculate the new positions of points along their normals
        shift = value / 10
        new_positions = self.get_point_set() + (self.normals * shift)
        # Update the points layer with new positions
        new_colors = self.get_point_colors(new_positions)

        vol_shape = self.volume.shape

        new_positions[:, 0] = np.clip(new_positions[:, 0], 0, vol_shape[0] - 1)
        new_positions[:, 1] = np.clip(new_positions[:, 1], 0, vol_shape[1] - 1)
        new_positions[:, 2] = np.clip(new_positions[:, 2], 0, vol_shape[2] - 1)

        self.color_values = new_colors
        self.vertices = new_positions
        self.update_mesh()
        self.slider_value.setText(f"{shift} vx")

    def update_mesh(self):
        self.surface_layer.data = (
            self.vertices,
            self.get_faces(),
            self.color_values,
        )

    def update_colors_based_on_sampling(self, value):
        spacing = 0.5
        sampling_depth = value / 10

        # Collect all samples for normalization calculation
        all_samples = []

        # Sample along the normal for each point
        for point, normal in zip(self.get_point_set(), self.normals):
            for depth in range(int(sampling_depth)):
                sample_point = point + normal * spacing * depth
                sample_point_clipped = np.clip(
                    sample_point, [0, 0, 0], np.array(self.volume.shape) - 1
                ).astype(int)
                sample_value = self.volume[
                    sample_point_clipped[0],
                    sample_point_clipped[1],
                    sample_point_clipped[2],
                ]
                all_samples.append(sample_value)

        # Calculate min and max across all sampled values
        samples_min = np.min(all_samples)
        samples_max = np.max(all_samples)

        # Normalize and update colors based on the mean value
        # of samples for each point
        new_colors = np.zeros((len(self.get_point_set()),))
        for i, (point, normal) in enumerate(
            zip(self.get_point_set(), self.normals)
        ):
            samples = []
            for depth in range(int(sampling_depth)):
                sample_point = point + normal * spacing * depth
                sample_point_clipped = np.clip(
                    sample_point, [0, 0, 0], np.array(self.volume.shape) - 1
                ).astype(int)
                sample_value = self.volume[
                    sample_point_clipped[0],
                    sample_point_clipped[1],
                    sample_point_clipped[2],
                ]
                samples.append(sample_value)

            # Normalize the mean of samples for this point using
            # the min and max from all samples
            mean_value = np.mean(samples)
            normalized_value = (
                (mean_value - samples_min) / (samples_max - samples_min)
                if samples_max > samples_min
                else 0
            )
            new_colors[i] = normalized_value

        self.color_values = new_colors
        self.update_mesh()
        self.sampling_depth_value.setText(f"{value}")


if __name__ == "__main__":

    obj_path = "../../examples/tomo_17_M10_grow1_1_mesh_data.obj"
    tomo_path = "../../examples/tomo_17_M10_grow1_1_mesh_data.mrc"

    mrc = mrcfile.open(tomo_path, permissive=True)
    tomo_mrc = np.array(mrc.data)

    vertices, faces, values = read_obj_file(obj_path)
    surface = (vertices, faces, values)

    viewer = napari.Viewer(ndisplay=3)
    volume_layer = viewer.add_image(tomo_mrc)
    surface_layer = viewer.add_surface(surface)

    # Testing points

    point_set = surface[0]

    volume_shape = np.array(tomo_mrc.data.shape)
    points_indices = np.round(point_set).astype(int)

    # Instantiate the widget and add it to Napari
    surforama_widget = QtSurforama(viewer, surface_layer, volume_layer)
    viewer.window.add_dock_widget(
        surforama_widget, area="right", name="Surforama"
    )

    napari.run()



from typing import List, Optional

import napari
from magicgui import magicgui
from napari.layers import Labels
from qtpy.QtWidgets import QGroupBox, QVBoxLayout, QWidget

from surforama.io import convert_mask_to_mesh


class QtMeshGenerator(QGroupBox):
    def __init__(
        self, viewer: napari.Viewer, parent: Optional[QWidget] = None
    ):
        super().__init__("Generate Mesh from Labels", parent=parent)
        self.viewer = viewer

        # make the labels layer selection widget
        self.labels_layer_selection_widget = magicgui(
            self._generate_mesh_from_labels,
            labels_layer={"choices": self._get_valid_labels_layers},
            barycentric_area={
                "widget_type": "Slider",
                "min": 0.1,
                "max": 10.0,
                "value": 1.0,
                "step": 0.1,
            },
            smoothing={
                "widget_type": "Slider",
                "min": 0,
                "max": 1000,
                "value": 1000,
            },
            call_button="Generate Mesh",
        )

        # make the layout
        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.labels_layer_selection_widget.native)

        # Add callback to update choices when layers change
        self.viewer.layers.events.inserted.connect(self._on_layer_update)
        self.viewer.layers.events.removed.connect(self._on_layer_update)

    def _on_layer_update(self, event=None):
        """Refresh the layer choices when layers are added or removed."""
        self.labels_layer_selection_widget.reset_choices()

    def _get_valid_labels_layers(self, combo_box) -> List[Labels]:
        return [
            layer
            for layer in self.viewer.layers
            if isinstance(layer, napari.layers.Labels)
        ]

    def _generate_mesh_from_labels(
        self,
        labels_layer: Labels,
        smoothing: int = 10,
        barycentric_area: float = 1.0,
    ):
        # Assuming create_mesh_from_mask exists and generates vertices, faces, and values
        vertices, faces, values = convert_mask_to_mesh(
            labels_layer.data,
            smoothing=smoothing,
            barycentric_area=barycentric_area,
        )
        self.viewer.add_surface((vertices, faces, values))



from typing import TYPE_CHECKING, Optional

import napari
import numpy as np
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QGroupBox,
    QPushButton,
    QVBoxLayout,
    QWidget,
)
from superqt import QLabeledDoubleSlider

if TYPE_CHECKING:
    from surforama import QtSurforama
from surforama.constants import (
    NAPARI_NORMAL_0,
    NAPARI_NORMAL_1,
    NAPARI_NORMAL_2,
    NAPARI_UP_0,
    NAPARI_UP_1,
    NAPARI_UP_2,
    ROTATION,
)
from surforama.utils.geometry import rotate_around_vector
from surforama.utils.napari import (
    update_rotations_on_points_layer,
    vectors_data_from_points_layer,
)


class QtSurfacePicker(QGroupBox):
    ENABLE_BUTTON_TEXT = "Enable"
    DISABLE_BUTTON_TEXT = "Disable"

    def __init__(
        self, surforama: "QtSurforama", parent: Optional[QWidget] = None
    ):
        super().__init__("Pick on surface", parent=parent)
        self.surforama = surforama
        self.points_layer = None
        self.normal_vectors_layer = None
        self.up_vectors_layer = None

        self.surforama.viewer.layers.events.removed.connect(
            self._on_layer_update
        )

        # initialize orientation data
        # todo store elsewhere (e.g., layer features)
        self.normal_vectors = np.empty((0, 3))
        self.up_vectors = np.empty((0, 3))
        self.rotations = np.empty((0,))

        # enable state
        self._enabled = False

        # make the activate button
        self.enable_button = QPushButton(self.ENABLE_BUTTON_TEXT)
        self.enable_button.clicked.connect(self._on_enable_button_pressed)

        # make the rotation slider
        self.rotation_slider = QLabeledDoubleSlider(Qt.Orientation.Horizontal)
        self.rotation_slider.setMinimum(-180)
        self.rotation_slider.setMaximum(180)
        self.rotation_slider.setValue(0)
        self.rotation_slider.valueChanged.connect(self._update_rotation)

        # make the layout
        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.enable_button)
        self.layout().addWidget(self.rotation_slider)

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, enabled: bool):
        if enabled == self.enabled:
            # do nothing
            return
        if enabled:
            self.enable_button.setText(self.DISABLE_BUTTON_TEXT)

            if self.points_layer is None:
                self._initialize_points_layer()
            if self.normal_vectors_layer is None:
                self._initialize_normal_vectors_layers()
            if self.up_vectors_layer is None:
                self._initialize_up_vectors_layers()
            self.points_layer.visible = True

            # update the vectors layer
            self._on_points_update()

            # add the mouse callbacks
            self._connect_mouse_callbacks()

        else:
            self.enable_button.setText(self.ENABLE_BUTTON_TEXT)

            # remove the mouse callbacks
            self._disconnect_mouse_callbacks()

        self._enabled = enabled

    def _on_layer_update(self):
        # check if the stored layers are still around
        viewer = self.surforama.viewer
        layer_deleted = False
        if (self.points_layer is not None) and (
            self.points_layer not in viewer.layers
        ):
            # remove the surface layer if it has been deleted.
            self.points_layer = None
            layer_deleted = True

        if (self.normal_vectors_layer is not None) and (
            self.normal_vectors_layer not in viewer.layers
        ):
            # remove the surface layer if it has been deleted.
            self.normal_vectors_layer = None
            layer_deleted = True

        if (self.up_vectors_layer is not None) and (
            self.up_vectors_layer not in viewer.layers
        ):
            # remove the surface layer if it has been deleted.
            self.up_vectors_layer = None
            layer_deleted = True

        if layer_deleted:
            self.enabled = False

    def _on_enable_button_pressed(self, event):
        # toggle enabled
        if self.enabled:
            # if currently enabled, toggle to disabled
            self.enabled = False

        else:
            # if disabled, toggle to enabled
            self.enabled = True

    def _update_rotation(self, value):
        """Callback function to update the rotation of the selected points."""
        selected_points = list(self.points_layer.selected_data)
        self.rotations[selected_points] = value

        rotation_radians = value * (np.pi / 180)
        new_rotations = rotation_radians * np.ones(
            len(selected_points), dtype=float
        )

        old_up_vector = self.up_vectors[selected_points]
        normal_vector = self.normal_vectors[selected_points]

        new_up_vector = rotate_around_vector(
            rotate_around=normal_vector,
            to_rotate=old_up_vector,
            angle=rotation_radians,
        )

        update_rotations_on_points_layer(
            points_layer=self.points_layer,
            point_index=selected_points,
            rotations=new_rotations,
        )

        self.up_vectors_layer.data[selected_points, 1, :] = new_up_vector
        self.up_vectors_layer.refresh()

    def _initialize_points_layer(self):
        self.points_layer = self.surforama.viewer.add_points(
            ndim=3, size=3, face_color="magenta"
        )
        self.points_layer.shading = "spherical"
        self.points_layer.events.data.connect(self._on_points_update)
        self.points_layer.selected_data.events.items_changed.connect(
            self._on_point_selection
        )
        self.surforama.viewer.layers.selection = [self.surforama.surface_layer]

    def _initialize_normal_vectors_layers(self):
        self.normal_vectors_layer = self.surforama.viewer.add_vectors(
            ndim=3,
            length=10,
            edge_color="cornflowerblue",
            name="surface normals",
        )
        self.surforama.viewer.layers.selection = [self.surforama.surface_layer]

    def _initialize_up_vectors_layers(self):
        self.up_vectors_layer = self.surforama.viewer.add_vectors(
            ndim=3, length=10, edge_color="orange", name="up vectors"
        )
        self.surforama.viewer.layers.selection = [self.surforama.surface_layer]

    def _on_points_update(self, event=None):
        """Update the vectors layers when the points data are updated."""
        # update the vectors
        normal_data, up_data = vectors_data_from_points_layer(
            self.points_layer
        )
        self.normal_vectors_layer.data = normal_data
        self.up_vectors_layer.data = up_data

        # colors were being reset - this might not be necessary
        self.normal_vectors_layer.edge_color = "purple"
        self.up_vectors_layer.edge_color = "orange"

    def _on_point_selection(self, event=None):
        selected_points = list(self.points_layer.selected_data)
        if len(selected_points) == 0:
            # no points selected
            return
        rotation_column = self.points_layer.features.columns.get_loc(ROTATION)
        rotations = self.points_layer.features.iloc[
            selected_points, rotation_column
        ].to_numpy()
        rotation = rotations[0]

        self.rotation_slider.blockSignals(True)
        self.rotation_slider.setValue((180 / np.pi) * rotation)
        self.rotation_slider.blockSignals(False)

    def _connect_mouse_callbacks(self):
        self.surforama.surface_layer.mouse_drag_callbacks.append(
            self._find_point_on_click
        )

    def _disconnect_mouse_callbacks(self):
        self.surforama.surface_layer.mouse_drag_callbacks.remove(
            self._find_point_on_click
        )

    def _find_point_on_click(self, layer, event):
        # if "Alt" not in event.modifiers:
        #    return
        value = layer.get_value(
            event.position,
            view_direction=event.view_direction,
            dims_displayed=event.dims_displayed,
            world=True,
        )
        if value is None:
            return
        triangle_index = value[1]
        if triangle_index is None:
            # if the click did not intersect the mesh, don't do anything
            return

        # get the intersection point
        candidate_vertices = layer.data[1][triangle_index]
        candidate_points = layer.data[0][candidate_vertices]
        (
            _,
            intersection_coords,
        ) = napari.utils.geometry.find_nearest_triangle_intersection(
            event.position, event.view_direction, candidate_points[None, :, :]
        )

        # get normal vector of intersected triangle
        mesh = self.surforama.mesh
        normal_vector = mesh.face_normals[triangle_index]

        # create the orientation coordinate system
        up_vector = np.cross(
            normal_vector, [1, 0, 0]
        )  # todo add check if normal is parallel

        # store the data
        feature_table = self.points_layer._feature_table
        table_defaults = feature_table.defaults
        table_defaults[NAPARI_NORMAL_0] = normal_vector[0]
        table_defaults[NAPARI_NORMAL_1] = normal_vector[1]
        table_defaults[NAPARI_NORMAL_2] = normal_vector[2]
        table_defaults[NAPARI_UP_0] = up_vector[0]
        table_defaults[NAPARI_UP_1] = up_vector[1]
        table_defaults[NAPARI_UP_2] = up_vector[2]
        table_defaults[ROTATION] = 0.0
        self.normal_vectors = np.concatenate(
            (self.normal_vectors, np.atleast_2d(normal_vector))
        )
        self.up_vectors = np.concatenate(
            (self.up_vectors, np.atleast_2d(up_vector))
        )
        self.rotations = np.append(self.rotations, 0)

        with self.points_layer.events.data.blocker(self._on_points_update):
            # we block since the event emission is before the features are updated.
            self.points_layer.add(np.atleast_2d(intersection_coords))
        self._on_points_update()



import numpy as np
import potpourri3d as pp3d
import pyvista as pv
from pygeodesic import geodesic

from surforama.constants import NAPARI_UP_0, NAPARI_UP_1, NAPARI_UP_2


def find_closest_vertex(vertices, point):
    """Find the index of the closest vertex to a given point."""
    distances = np.linalg.norm(vertices - point, axis=1)
    return np.argmin(distances)


def compute_geod_distance_matrix(
    verts, faces, point_coordinates, method="exact"
):
    """
    Compute the geodesic distance matrix between a set of points on a mesh.

    Has two methods for computing the geodesic distances:
    - "exact": Uses the exact geodesic algorithm from the pygeodesic library.
        (takes forever for large meshes and many points)
    - "fast": Uses the mesh heat method distance solver from the potpourri3d library.
        (faster but less accurate -- but still good enough for most applications;
        differences are in the area of ~[-0.2, 0.2] vx for the distances between points on the same mesh)

    Parameters
    ----------
    verts : np.ndarray
        The vertices of the mesh.
    faces : np.ndarray
        The faces of the mesh.
    point_coordinates : np.ndarray
        The coordinates of the points for which the geodesic distances should be computed.
    method : str
        The method to use for computing the geodesic distances. Can be either "exact" or "fast".

    Returns
    -------
    np.ndarray
        The geodesic distance matrix between the points.

    """

    if method == "exact":
        geoalg = geodesic.PyGeodesicAlgorithmExact(verts, faces)
    elif method == "fast":
        solver = pp3d.MeshHeatMethodDistanceSolver(V=verts, F=faces)

    point_idcs = [
        find_closest_vertex(verts, point) for point in point_coordinates
    ]

    distance_matrix = np.zeros((len(point_idcs), len(point_idcs))).astype(
        np.float32
    )
    for i, point_idx in enumerate(point_idcs):
        if method == "exact":
            distances, _ = geoalg.geodesicDistances(
                np.array([point_idx]), np.arange(len(verts))
            )
        elif method == "fast":
            distances = solver.compute_distance(point_idx)

        distances[point_idx] = 1e5
        distances = distances[point_idcs]
        distance_matrix[i] = distances

    return distance_matrix


def compute_surface_occupancy(
    verts, faces, point_coordinates, only_front=True
):
    """
    Compute the surface occupancy of a set of points on a mesh.

    The surface occupancy is defined as the number of points divided by the surface area
    of the mesh that is covered by the points.

    Parameters
    ----------
    verts : np.ndarray
        The vertices of the mesh.
    faces : np.ndarray
        The faces of the mesh.
    point_coordinates : np.ndarray
        The coordinates of the points for which the surface occupancy should be computed.
    only_front : bool
        If True, only the front side of the mesh is considered for the surface area computation.
        This means that the surface area is divided by 2.

    Returns
    -------
    float
        The surface occupancy of the points on the mesh.
    """
    pv_faces = np.hstack([np.ones((faces.shape[0], 1)) * 3, faces])
    pv_faces = pv_faces.flatten().astype(np.int32)
    pv_mesh = pv.PolyData(verts, pv_faces)
    area = pv_mesh.area
    if only_front:
        area /= 2.0  # divide by 2 to only take the front side
    return len(point_coordinates) / area


def orientations_of_knn_inplane(
    distance_matrix, feature_table, k=3, c2_symmetry=False
):
    """
    Compute the angular differences between the up vectors of a point and its k nearest neighbors.

    The angular differences are computed in degrees and are in the range [0, 180] if c2_symmetry is False
    and in the range [0, 90] if c2_symmetry is True.

    Parameters
    ----------
    distance_matrix : np.ndarray
        The distance matrix between the points.
    feature_table : pd.DataFrame
        The feature table containing the up vectors of the points.
    k : int
        The number of nearest neighbors to consider.
    c2_symmetry : bool
        Whether to consider the c2 symmetry of the up vectors.
        (implies C2 symmetry of the respective protein)

    Returns
    -------
    list
        A list of angular differences for each point. (in degrees)
    """
    up_vectors = feature_table[
        [NAPARI_UP_0, NAPARI_UP_1, NAPARI_UP_2]
    ].to_numpy()
    nn_orientations = []
    for i in range(distance_matrix.shape[0]):
        start_vector = up_vectors[i]
        nn_idx = np.argsort(distance_matrix[i])[:k]
        cosine_similarities = np.degrees(
            np.arccos(np.dot(start_vector, up_vectors[nn_idx].T))
        )
        if c2_symmetry:
            cosine_similarities = np.minimum(
                cosine_similarities, 180 - cosine_similarities
            )
        nn_orientations.append(cosine_similarities)
    return np.array(nn_orientations)



from typing import Callable, List

from napari.types import LayerDataTuple

from surforama.io import read_obj_file
from surforama.io.star import load_points_layer_data_from_star_file
from surforama.utils.napari import vectors_data_from_points_data


def mesh_reader_plugin(path: str) -> Callable[[str], LayerDataTuple]:
    if isinstance(path, str) and path.endswith(".obj"):
        return obj_reader

    # format not recognized
    return None


def obj_reader(path: str) -> List[LayerDataTuple]:
    mesh_data = read_obj_file(path)
    return [(mesh_data, {}, "surface")]


def star_reader_plugin(path: str) -> Callable[[str], LayerDataTuple]:
    if isinstance(path, str) and path.endswith(".star"):
        return star_reader

    # format not recognized
    return None


def star_reader(path: str) -> List[LayerDataTuple]:
    point_coordinates, features_table = load_points_layer_data_from_star_file(
        file_path=path
    )

    # make the points layer data
    points_layer_data = (
        point_coordinates,
        {"size": 3, "face_color": "magenta", "features": features_table},
        "points",
    )

    # make the vectors layer data
    normal_data, up_data = vectors_data_from_points_data(
        point_coordinates=point_coordinates, features_table=features_table
    )
    normal_layer_data = (
        normal_data,
        {"length": 10, "edge_color": "orange", "name": "normal vectors"},
        "vectors",
    )
    up_layer_data = (
        up_data,
        {"length": 10, "edge_color": "purple", "name": "up vectors"},
        "vectors",
    )

    return [points_layer_data, normal_layer_data, up_layer_data]



from typing import Union

import numpy as np
from scipy.spatial.transform import Rotation as R


def rotate_around_vector(
    rotate_around: np.ndarray,
    to_rotate: np.ndarray,
    angle: Union[float, np.ndarray],
) -> np.ndarray:
    """Rotate a vector around another vector a specified angle.

    Parameters
    ----------
    rotate_around : np.ndarray
        The vector to rotate around.
    to_rotate : np.ndarray
        The vector to rotate.
    angle : Union[float, np.ndarray]
        The angle to perform the rotation in radians.
        Positive angles will rotate counter-clockwise.
        If provided as an array, each element should be indexed-matched
        with the vectors.

    Returns
    -------
    np.ndarray
        The rotated vector.
    """
    # ensure unit vectors
    rotate_around = rotate_around.astype(float)
    if rotate_around.ndim == 1:
        rotate_around /= np.linalg.norm(rotate_around)
    elif rotate_around.ndim == 2:
        # handle the 2D case
        rotate_around /= np.linalg.norm(rotate_around, axis=1)[..., np.newaxis]
        angle = np.asarray(angle)[..., np.newaxis]
    else:
        raise ValueError(
            "rotate_around must be a single vector as a 1D array or a 2D array of multiple vectors."
        )

    rotation_vector = angle * rotate_around
    rotation_object = R.from_rotvec(rotation_vector)

    return rotation_object.apply(to_rotate)



import numpy as np
import pyacvd
import pyvista as pv
import trimesh
from skimage.measure import marching_cubes


def read_obj_file(file_path):
    mesh = trimesh.load(file_path, file_type="obj", process=True)

    # Subdivide
    # verts, faces = trimesh.remesh.subdivide_to_size(
    # mesh.vertices, mesh.faces, 1
    # )

    # Subdivide can introduce holes
    # mesh = trimesh.Trimesh(vertices=verts, faces=faces)
    # trimesh.repair.fill_holes(mesh)

    verts = mesh.vertices
    faces = mesh.faces

    # trimesh swaps coordinate axes
    verts = verts[:, [2, 1, 0]]

    values = np.ones((len(verts),))

    return verts, faces, values


def convert_mask_to_mesh(
    mask: np.ndarray,
    barycentric_area: float = 1.0,
    smoothing: int = 10,
):
    """
    Convert a binary mask to a mesh.

    Parameters
    ----------
    mask : np.ndarray
        A binary mask.
    barycentric_area : float, optional
        The target barycentric area of each vertex in the mesh,
        by default 1.0
    smoothing : int, optional
        Number of iterations for Laplacian mesh smoothing,
        by default 10
    """
    verts, faces, _, _ = marching_cubes(
        volume=mask,
        level=0.5,
        step_size=1,
        method="lewiner",
    )

    # Prepend 3 for pyvista format
    faces = np.concatenate(
        (np.ones((faces.shape[0], 1), dtype=int) * 3, faces), axis=1
    )

    # Create a mesh
    surf = pv.PolyData(verts, faces)
    surf = surf.smooth(n_iter=smoothing)

    # remesh to desired point size
    cluster_points = int(surf.area / barycentric_area)
    clus = pyacvd.Clustering(surf)
    clus.subdivide(3)
    clus.cluster(cluster_points)
    remesh = clus.create_mesh()

    verts = remesh.points
    faces = remesh.faces.reshape(-1, 4)[:, 1:]
    values = np.ones((len(verts),))

    # switch face order to have outward normals
    faces = faces[:, [0, 2, 1]]

    return verts, faces, values
